import asyncio
import logging

from .protocol import SyncProtocol
from .processor import ProcessorRegistry

class ManagerRegistry(type):
  REGISTRY = {}

  def __new__(mcs, name, bases, attrs):
    """ @param name: Name of the class
        @param bases: Base classes (tuple)
        @param attrs: Attributes defined for the class
    """
    new_cls = type.__new__(mcs, name, bases, attrs)
    mcs.REGISTRY[new_cls.name] = new_cls

    return new_cls


class Manager(metaclass = ManagerRegistry):
  name = "default_manager"
  source = None
  run_always = True

  def __init__(self, handler = SyncProtocol()):
    self.handler = handler
    self.logger = logging.getLogger(self.name)

  def build_processor_name(self, name):
    return '.'.join([self.name, name])

  def get_source(self):
    return self.source

  def done(self):
    if self.source:
      return self.source.is_done()

  def should_restart(self):
    return self.run_always

  def start(self, loop):
    asyncio.ensure_future(self.get_source().start(), loop = loop)

  def remove_all_listeners(self):
    """
    Remove all listeners attached to self.
    """
    events = [event for event in self.handler.get_events() if event.startswith(self.name + '.')]

    for event in events:
      self.handler.remove_all_listeners(event)

class DescriptiveManager(Manager, metaclass = ManagerRegistry):
  processors = {}
  flows = []

  def __init__(self, handler = SyncProtocol()):
    super().__init__(handler)
    for name, metadata in self.processors.items():
      processor = ProcessorRegistry.REGISTRY[metadata['cls']](self.build_processor_name(name), self.handler, *metadata['args'], **metadata['kwargs'])
      setattr(self, name, processor)

    for flow in self.flows:
      origin = getattr(self, flow['origin'])
      target = getattr(self, flow['target'])
      target_function = getattr(target, flow['function'])
      origin.on(flow['event'], target_function)



#pylint: disable=E0401
"""
Test Pusher client source.
"""
import asyncio
from unittest import mock
from unittest import TestCase

from aiopype.sources import PusherClientSource


class TestPusherClientSource(TestCase):
  def test_start(self):
    mock_handler = mock.MagicMock()
    source = PusherClientSource('test', mock_handler)
    source.pusher = mock.MagicMock()
    type(source.pusher).exception = mock.PropertyMock(side_effect = [None, 'a', Exception('forcequit')])

    loop = asyncio.get_event_loop()

    try:
      loop.run_until_complete(source.start())
    except Exception as error:
      self.assertEqual(str(error), 'forcequit')

    self.assertTrue(source.done)

    type(source.pusher).exception = mock.PropertyMock(return_value = None)
    type(source.pusher.connection).state = mock.PropertyMock(return_value = 'failed')

    try:
      loop.run_until_complete(source.start())
    except Exception as error:
      self.assertEqual(str(error), 'Connection to pusherclient lost failed')



import asyncio
import json
import logging
from datetime import datetime, timedelta

import aiohttp
import websockets

from aiopype.processor import Processor

PACKET_NAMES = {
  '0': 'disconnect',
  '1': 'connect',
  '2': 'heartbeat',
  '3': 'message',
  '4': 'json',
  '5': 'event',
  '6': 'ack',
  '7': 'error',
  '8': 'noop',
}

PACKET_CODES = {v: k for k, v in PACKET_NAMES.items()}


class SocketIOPacketException(Exception):
  pass


class SocketIOPacket(object):
  def __init__(self, packet_type, name, data):
    self.data = data
    self.name = name
    self.packet_type = packet_type
    self.packet_code = PACKET_CODES[self.packet_type]

  @classmethod
  def parse(cls, packet):
    parts = packet.split(':')
    return cls(PACKET_NAMES[parts[0]], None, ':'.join(parts[3:]))

  def __str__(self):
    parts = [
      self.packet_code,
      '',
      ''
    ]

    if self.packet_code == '5':
      parts.append(json.dumps({
        'name': self.name,
        'args': [self.data]
      }))

    return ':'.join(parts)


class SocketIOSource(Processor):
  response_factory = SocketIOPacket
  source = 'localhost'

  def __init__(self, name, handler, *args, **kwargs):
    super().__init__(name, handler)
    self.done = None
    self.failures = 0
    self.last_heartbeat = datetime.now()
    self.logger = logging.getLogger(name)
    self.loop = asyncio.get_event_loop()
    self.max_failures = kwargs.get('exception_threshold', 10)
    self.heartbeat_timeout = kwargs.get('heartbeat_timeout', 60)
    self.on('connected', self.on_connect)
    self.on('disconnect', self.on_disconnect)
    self.on('error', self.on_error)
    self.on('heartbeat', self.handle_heartbeat)
    self.running = True
    self.websocket = None

  async def on_connect(self):
    pass

  async def on_error(self, _):
    self.websocket.close()
    self.logger.warning('Got unexpected error')
    raise websockets.ConnectionClosed(None, None)

  async def on_disconnect(self, _):
    self.logger.warning("Disconnected")
    self.failures += 1
    if self.failures > self.max_failures:
      self.running = False

  async def get_url(self):
    with aiohttp.ClientSession() as session:
      response = await session.post('http://{}/socket.io/1/'.format(self.source))
      data = await response.text()
      hskey = data.split(':')[0]

    return 'ws://{}/socket.io/1/websocket/{}'.format(self.source, hskey)

  async def start(self):
    try:
      while self.running:
        self.last_heartbeat = datetime.now()
        await self.watchdog(self.connect())
    finally:
      self.done = True
      raise Exception("Disconnected too many times, stopped")

  async def connect(self):
    try:
      url = await self.get_url()

      async with websockets.connect(url) as self.websocket:
        self.logger.debug('Connected to websocket {}'.format(url))
        await self.emit_async('connected')

        while self.running:
          data_raw = await self.websocket.recv()
          received = self.response_factory.parse(data_raw)
          await self.emit_async(received.packet_type, received)
          self.failures = 0

    except (websockets.ConnectionClosed, ConnectionResetError, websockets.InvalidHandshake):
      await self.emit_async('disconnect', "Websocket closed")

  async def watchdog(self, coro):
    future = asyncio.ensure_future(coro, loop = self.loop)

    while not future.done() and datetime.now() - self.last_heartbeat < timedelta(seconds = self.heartbeat_timeout):
      await asyncio.sleep(1)

    if not future.done():
      future.cancel()

    await self.emit_async('disconnect', "Lost heartbeat")

  async def handle_heartbeat(self, _):
    self.last_heartbeat = datetime.now()
    request = SocketIOPacket('heartbeat', '', '')
    await self.websocket.send(str(request))



import asyncio
from collections import defaultdict
from itertools import chain

from .sync_protocol import SyncProtocol


class AsyncProtocol(SyncProtocol):
  def __init__(self):
    super().__init__()
    self.running = False
    self.queue = asyncio.Queue()
    self.loop = asyncio.get_event_loop()
    self.on('stop', self.stop)

  def emit(self, event, *args, **kwargs):
    asyncio.ensure_future(self.queue.put((event, args, kwargs)), loop = self.loop)
    return True

  async def emit_async(self, event, *args, **kwargs):
    await self.queue.put((event, args, kwargs))
    return True

  def stop(self):
    self.running = False

  async def listen(self):
    self.running = True
    self.tasks = defaultdict(list)

    try:
      while self.running:
        event, args, kwargs = await self.queue.get()

        if not self.tasks[event]:
          handled = False
          events_copy = list(self._events[event])

          # Pass the args to each function in the events dict
          for handler in events_copy:
            if asyncio.iscoroutinefunction(handler):
              self.tasks[event].append(asyncio.ensure_future(handler(*args, **kwargs)))
            else:
              handler(*args, **kwargs)
            handled = True

          if not handled and event.split('__')[-1] == 'error':
            raise Exception('Uncaught error event.')
        else:
          self.queue.put((event, args, kwargs))

        for event, tasks in self.tasks.items():
          self.tasks[event] = [task for task in tasks if not task.done()]

    finally:
      self.running = False
      await asyncio.wait_for(asyncio.gather(*chain(*self.tasks.values()), return_exceptions = True), 10)



#pylint: disable=E0401
"""
Test flow controller.
"""
from unittest import mock
from unittest import TestCase

from aiopype import SyncProtocol
from aiopype.manager import Manager, DescriptiveManager


class TestManager(TestCase):
  def test_get_source(self):
    manager = Manager()
    manager.source = "Test"

    source = manager.get_source()

    self.assertEqual("Test", source)

  def test_done(self):
    manager = Manager()
    manager.source = mock.Mock()
    manager.source.is_done = mock.Mock(return_value = False)

    self.assertFalse(manager.done())

    manager.source.is_done = mock.Mock(return_value = True)

    self.assertTrue(manager.done())

  def test_restart(self):
    manager = Manager()
    manager.run_always = False

    self.assertFalse(manager.should_restart())

    manager.run_always = True

    self.assertTrue(manager.should_restart())

  @mock.patch('aiopype.manager.asyncio')
  def test_start(self, asyncio_mock):
    asyncio_mock.ensure_future = mock.Mock()
    manager = Manager()
    manager.source = mock.Mock()

    manager.source.start = mock.Mock(return_value = 'Test')
    loop = 'loop'

    manager.start(loop)

    asyncio_mock.ensure_future.assert_called_with('Test', loop = 'loop')


class TestDescriptiveManager(TestCase):
  @mock.patch('aiopype.manager.ProcessorRegistry')
  def test_init(self, registry_mock):
    mocksource = mock.MagicMock()
    registry_mock.REGISTRY = {
      'mocksource': mocksource
    }

    class MockDescriptiveManager(DescriptiveManager):
      name = 'mockmanager'
      processors = {
        'source' : {
          'cls': 'mocksource',
          'args': ['a', 'b'],
          'kwargs': {'c' : 1}
        }
      }
      flows = [
        {
          'event': 'mockevent',
          'function': 'test',
          'origin': 'source',
          'target': 'source',
        }
      ]
    handler = SyncProtocol()
    mock_manager = MockDescriptiveManager(handler = handler)

    self.assertTrue(mock_manager.source)
    mocksource.assert_called_with('mockmanager.source', handler, 'a', 'b', c = 1)
    mock_manager.source.on.assert_called_with('mockevent', mock_manager.source.test)



"""
Time Sources.
"""
import asyncio
from datetime import datetime

from aiopype.processor import Processor


def get_rounded_timestamp(interval):
  time = datetime.now().timestamp()
  time -= time % interval
  return time


class TimeSource(Processor):
  def __init__(self, name, handler, interval):
    super().__init__(name, handler)
    self.interval = interval
    self.loop = asyncio.get_event_loop()

  def get_time(self):
    return get_rounded_timestamp(self.interval)

  async def start(self):
    await self.emit_async('cron', self.get_time())
    self.done = True


class OffsetTimeSource(TimeSource):
  def __init__(self, name, handler, interval, offset):
    super().__init__(name, handler, interval)
    self.offset = offset

  def get_time(self):
    return get_rounded_timestamp(self.interval) - self.offset



#pylint: disable=E0401
"""
Test websocket source.
"""
import asyncio
from unittest import mock
from unittest import TestCase

from aiopype import SyncProtocol
from aiopype.sources import WebsocketSource


class TestWebsocketSource(TestCase):
  @mock.patch('aiopype.sources.WebsocketSource.on_connect')
  @mock.patch('aiopype.sources.WebsocketSource.on_lost_connection')
  def test_init(self, lost_connection_mock, connected_mock):
    source = WebsocketSource('test', SyncProtocol())
    done_future = asyncio.Future()
    done_future.set_result(None)
    connected_mock.return_value = done_future
    lost_connection_mock.return_value = done_future

    async def test_events():
      await source.emit_async('connected')
      await source.emit_async('disconnected')

    loop = asyncio.get_event_loop()
    loop.run_until_complete(test_events())

    connected_mock.assert_called_with()
    lost_connection_mock.assert_called_with()
    self.assertEqual(source.done, None)
    self.assertEqual(source.failures, 0)
    self.assertEqual(source.max_failures, 10)
    self.assertEqual(source.heartbeat_timeout, 30)

  def test_lost_connection(self):
    source = WebsocketSource('test', SyncProtocol())
    source.max_failures = 0
    loop = asyncio.get_event_loop()
    loop.run_until_complete(source.on_lost_connection())
    self.assertFalse(source.running)

  def test_start(self):
    source = WebsocketSource('test', SyncProtocol())
    source.connect = mock.Mock(return_value = 'test')
    watch_future = asyncio.Future()
    watch_future.set_result(None)
    source.watchdog = mock.Mock(return_value = watch_future)
    WebsocketSource.running = mock.PropertyMock(side_effect = [True, True, False])

    loop = asyncio.get_event_loop()
    exception = None

    try:
      loop.run_until_complete(source.start())
    except Exception as error:
      exception = error
    self.assertEqual(str(exception), 'Disconnected too many times, stopped')
    self.assertTrue(source.done)
    source.watchdog.assert_called_with('test')
    WebsocketSource.running = True

  def test_parse(self):
    parsed = WebsocketSource.parse("[1, 2, 3]")
    self.assertEqual(parsed, [1, 2, 3])


class TestWebsocketSourceWatchdog(TestCase):
  def setUp(self):
    self.source = WebsocketSource('test', SyncProtocol())
    self.loop = asyncio.get_event_loop()

    done_future = asyncio.Future()
    done_future.set_result(None)
    self.source.emit_async = mock.Mock(return_value = done_future)

  async def sleeper(self, time):
    await asyncio.sleep(time)

  def test_finish_before_websocket(self):
    self.loop.run_until_complete(self.source.watchdog(self.sleeper(1)))
    self.source.emit_async.assert_called_with('disconnected', 'Service failed')

  def test_timeout_before_websocket(self):
    self.source.heartbeat_timeout = 0.01
    self.loop.run_until_complete(self.source.watchdog(self.sleeper(1)))
    self.source.emit_async.assert_called_with('disconnected', 'Unable to connect')

  def test_coroutine_done(self):
    self.source.websocket = mock.Mock()
    pong_future = asyncio.Future()
    pong_future.set_result(asyncio.ensure_future(self.sleeper(1)))
    self.source.websocket.ping = mock.Mock(return_value = pong_future)

    self.loop.run_until_complete(self.source.watchdog(self.sleeper(1)))
    self.source.emit_async.assert_called_with('disconnected', 'Service failed')

  def test_lost_heartbeat(self):
    self.source.websocket = mock.Mock()
    self.source.heartbeat_timeout = 0.01
    pong_future = asyncio.Future()
    pong_future.set_result(asyncio.ensure_future(self.sleeper(10)))
    self.source.websocket.ping = mock.Mock(return_value = pong_future)

    self.loop.run_until_complete(self.source.watchdog(self.sleeper(1)))
    self.source.emit_async.assert_called_with('disconnected', 'Lost heartbeat')



"""
Collector for REST api endpoints.
"""
import asyncio
import logging
from datetime import datetime

import aiohttp
from aiopype import Processor


class RestSource(Processor):
  """
  Data source that relies on http polling to gather new data.
  """
  running = True

  def __init__(self, name, handler, url, options):
    super().__init__(name, handler)
    self.done = False
    self.logger = logging.getLogger(name)
    self.loop = asyncio.get_event_loop()
    self.options = options
    self.tasks = []
    self.tryouts = 0
    self.url = url

  async def start(self):
    """
    Main loop function.

    This loop runs until running is set to False, or an exception arises. While
    it is running, it constantly triggers http requests with an interval of
    `request_interval` seconds.
    Exceptions are not handled here so that the default handler can define
    the behavior on such events.
    """
    try:
      while self.running:
        task = asyncio.ensure_future(self.get_data(), loop = self.loop)
        self.tasks.append(task)
        await asyncio.sleep(self.options['request_interval'])
        done = [task for task in self.tasks if task.done()]
        self.tasks = [task for task in self.tasks if not task.done()]

        for task in done:
          await task

    finally:
      self.logger.debug("I have to wait for my tasks", extra = {'tasks': str(self.tasks)})

      try:
        await asyncio.wait_for(asyncio.gather(*self.tasks, return_exceptions = True), 10)
      finally:
        self.done = True

  def error_handler(self, error):
    self.logger.warning(error)

  async def get_data(self):
    """
    The method that is triggered every `request_interval` seconds and is Responsible
    for performing the HTTP request and emiting the obtained data.
    """
    try:
      with aiohttp.Timeout(8):
        response = await aiohttp.request('get', self.url)
        cur_time = datetime.strptime(response.headers['date'], '%a, %d %b %Y %H:%M:%S %Z').timestamp()
        data = await response.json()

      self.emit('data', data, int(cur_time))
      self.tryouts = 0

    except asyncio.TimeoutError:
      return

    except Exception as err:
      self.tryouts += 1

      if self.tryouts > self.options['exception_threshold']:
        self.error_handler(err)



"""
Main module

Contains the persister singleton.
"""
import asyncio
import logging
import sys

from raven import Client

from .manager import ManagerRegistry
from .protocol import AsyncProtocol, SyncProtocol


class Flow(object):
  """
  Persister entrypoint

  Responsible for orchestration, error handling and report.
  """
  def __init__(self, config):
    self.config = config
    self.logger = logging.getLogger('persister')
    if not hasattr(config, 'RAVEN_DSN'):
      config.RAVEN_DSN = None
    if config.RAVEN_DSN:
      self.raven = Client(config.RAVEN_DSN, release = config.VERSION)
    else:
      self.raven = None

    self.loop = asyncio.get_event_loop()
    self.loop.set_exception_handler(self.default_handler)

  def default_handler(self, _, data):
    """
    Entrypoint for event loop's unhandled exceptions.
    """
    if 'exception' in data:
      exception = data['exception']
      self.logger.warning(exception)
      self.handle_exception(exception)

  def handle_exception(self, exception):
    """
    Exception handler.

    Not all exceptions are caught if they are not awaited. Since there are some
    background jobs that will run in parallel with the main task, they might have to
    invoke this directly.
    """
    exception_info = (type(exception), exception, exception.__traceback__)

    if self.raven:
      self.raven.captureException(exception_info)
    else:
      self.logger.info('Unable to capture exception on raven', extra = {'info': exception_info})

    self.logger.debug("Reported to sentry, will restart manager")
    asyncio.ensure_future(self.check_managers(), loop = self.loop)

  async def check_managers(self):
    self.logger.debug("Checking for stopped managers")

    done_managers = [manager for manager in self.managers if manager.done()]

    self.logger.debug('Stopped: ' + str(done_managers))

    for manager in done_managers:

      if manager.should_restart():
        manager.remove_all_listeners()
        factory = ManagerRegistry.REGISTRY[manager.name]
        new_manager = factory(self.handler)
        new_manager.start(self.loop)
        self.managers.append(new_manager)

      self.managers.remove(manager)

  async def finished(self):
    while self.managers:
      await asyncio.sleep(4)
      self.managers = [manager for manager in self.managers if manager.should_restart() or not manager.done()]
    self.logger.debug("All managers are done")

  def start(self):
    self.managers = []
    self.handler = SyncProtocol()
    flows = self.config.FLOWS

    for manager_factory in [ManagerRegistry.REGISTRY[x] for x in flows]:
      manager = manager_factory(self.handler)
      manager.start(self.loop)
      self.managers.append(manager)

    self.loop.run_until_complete(self.finished())


class AsyncFlow(Flow):
  async def check_managers(self):
    self.logger.debug("Checking for stopped managers")

    done_managers = [manager for manager in self.managers if manager.done()]

    self.logger.debug('Stopped: ' + str(done_managers))

    for manager in done_managers:

      if manager.should_restart():
        manager.remove_all_listeners()
        factory = ManagerRegistry.REGISTRY[manager.name]
        new_manager = factory(self.handler)
        new_manager.start(self.loop)
        self.managers.append(new_manager)

      self.managers.remove(manager)

    if not self.handler.running:
      self.logger.error("Listener stopped, restarting.")
      self.handler_task = asyncio.ensure_future(self.check_listener())

  async def check_listener(self):
    try:
      await self.handler.listen()
    except Exception as error:
      self.handle_exception(error)

  async def finished(self):
    await super().finished()
    await self.handler.emit_async('stop')
    await self.handler_task
    self.logger.debug("Messaging handler done")

  def start(self):
    self.managers = []
    self.handler = AsyncProtocol()
    flows = self.config.FLOWS

    for manager_factory in [ManagerRegistry.REGISTRY[x] for x in flows]:
      manager = manager_factory(self.handler)
      manager.start(self.loop)
      self.managers.append(manager)

    self.handler_task = asyncio.ensure_future(self.check_listener(), loop = self.loop)
    self.loop.run_until_complete(self.finished())



from setuptools import find_packages
from distutils.core import setup

try:
  import pypandoc
  LONG_DESCRIPTION = pypandoc.convert('README.md', 'rst')
  LONG_DESCRIPTION += '\n' + pypandoc.convert('CHANGELOG.md', 'rst')
except(IOError, ImportError, OSError):
  LONG_DESCRIPTION = open('README.md').read()
  LONG_DESCRIPTION += open('CHANGELOG.md').read()

setup(
  author = 'Jorge Alpedrinha Ramos',
  author_email = 'python@uphold.com',
  description = 'AioPype - Flow based programming with asyncio',
  install_requires = [
    'aiohttp==0.21.2',
    'pusherclient==0.3.0',
    'pyee==1.0.2',
    'raven==5.11.0',
    'websocket-client==0.37.0',
    'websockets==3.0',
  ],
  license = 'LICENSE',
  long_description = LONG_DESCRIPTION,
  name = 'aiopype',
  package_data = { '': ['*.yml']},
  packages = find_packages(),
  url = 'https://www.github.com/uphold/aiopype/',
  version = '0.1.3',
)



import asyncio
from pyee import EventEmitter


class SyncProtocol(EventEmitter):
  async def emit_async(self, event, *args, **kwargs):
    """
    Emit asynchrounously.

    This method should be promoted to a Pull Request to pyee.
    """
    handled = False

    # Copy the events dict first. Avoids a bug if the events dict gets
    # changed in the middle of the following for loop.
    events_copy = list(self._events[event])

    # Pass the args to each function in the events dict
    for handler in events_copy:
      if asyncio.iscoroutinefunction(handler):
        await handler(*args, **kwargs)
      else:
        handler(*args, **kwargs)

      handled = True

    if not handled and event.split('__')[-1] == 'error':
      raise Exception('Uncaught error event.')

    return handled

  def get_events(self):
    return self._events.keys()

  def once(self, event, f = None):
    """
    Override to accomodate async handlers
    """
    def _once(handler):
      if asyncio.iscoroutinefunction(handler):
        async def new_handler(*args, **kwargs):
          await handler(*args, **kwargs)
          self.remove_listener(event, new_handler)

      else:
        def new_handler(*args, **kwargs):
          handler(*args, **kwargs)
          self.remove_listener(event, new_handler)

      return new_handler

    def _wrapper(handler):
      self.on(event, _once(handler))

      return handler

    if f is None:
      return _wrapper
    else:
      _wrapper(f)
